import json
import asyncio
import logging

from slimagents import Response

from slimagents.config import logger
from slimagents.core import Agent


async def process_and_print_streaming_response(response):
    content = ""
    last_sender = ""

    async for chunk in response:
        if isinstance(chunk, Response):
            return chunk
        
        if "sender" in chunk:
            last_sender = chunk["sender"]

        if "content" in chunk and chunk["content"] is not None:
            if not content and last_sender:
                print(f"\033[94m{last_sender}:\033[0m", end=" ", flush=True)
                last_sender = ""
            print(chunk["content"], end="", flush=True)
            content += chunk["content"]

        if "tool_calls" in chunk and chunk["tool_calls"] is not None:
            for tool_call in chunk["tool_calls"]:
                f = tool_call["function"]
                name = f["name"]
                if not name:
                    continue
                print(f"\033[94m{last_sender}: \033[95m{name}\033[0m()")

        if "delim" in chunk and chunk["delim"] == "end" and content:
            print()  # End of response message
            content = ""


def pretty_print_messages(messages) -> None:
    for message in messages:
        if message["role"] != "assistant":
            continue

        # print agent name in blue
        print(f"\033[94m{message['sender']}\033[0m:", end=" ")

        # print response, if any
        if message["content"]:
            print(message["content"])

        # print tool calls in purple, if any
        tool_calls = message.get("tool_calls") or []
        if len(tool_calls) > 1:
            print()
        for tool_call in tool_calls:
            f = tool_call["function"]
            name, args = f["name"], f["arguments"]
            arg_str = json.dumps(json.loads(args)).replace(":", "=")
            print(f"\033[95m{name}\033[0m({arg_str[1:-1]})")


def enable_logging(log_level: int = logging.ERROR) -> None:
    logging.basicConfig(
        level=logging.ERROR,
        format='%(asctime)s - %(levelname)s - %(name)s: %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    logger.setLevel(log_level)


async def run_demo_loop_async(agent: Agent, stream=False, log_level: int = None) -> None:
    if log_level is not None:
        enable_logging(log_level)

    print("Starting SlimAgents CLI 🪶")

    memory = []

    while True:
        user_input = input("\033[90mUser\033[0m: ")
        response = await agent.run(user_input, stream=stream, memory=memory, stream_response=True, stream_delimiters=True, stream_tokens=False, stream_tool_calls=True)
        if stream:
            response = await process_and_print_streaming_response(response)
        else:
            pretty_print_messages(response.memory_delta)

        agent = response.agent


def run_demo_loop(agent, stream=False, log_level: int = None):
    """Synchronous wrapper for run_demo_loop_async"""
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    loop.run_until_complete(run_demo_loop_async(agent, stream, log_level))



from .core import Agent, Response, ToolResult
from .repl import run_demo_loop, run_demo_loop_async
from importlib.metadata import version
from .config import logger

try:
    __version__ = version("slimagents")
except Exception:
    __version__ = "unknown"
    
logger.name = __name__

__all__ = ["Agent", "Response", "ToolResult", "run_demo_loop", "run_demo_loop_async", "logger"]



from collections import defaultdict
import copy
from textwrap import dedent
import litellm
from litellm.caching.caching import Cache

litellm.cache = Cache(type="disk", disk_cache_dir="./test_cache")

def merge_fields(target, source):
    for key, value in source.items():
        if isinstance(value, str):
            target[key] += value
        elif value is not None and isinstance(value, dict):
            merge_fields(target[key], value)

def merge_chunk_(final_response: dict, delta: dict) -> None:
    delta.pop("role", None)
    merge_fields(final_response, delta)

    tool_calls = delta.get("tool_calls")
    if tool_calls and len(tool_calls) > 0:
        tool_call = tool_calls[0]
        index = tool_call.pop("index")
        type_ = tool_call.pop("type", None)
        final_tool_call = final_response["tool_calls"][index]
        merge_fields(final_tool_call, tool_call)
        if type_ and not final_tool_call.get("type"):
            # type = "function" is always returned by LiteLLM in the delta. Bug?
            # This ensures that the type is only set once.
            final_tool_call["type"] = type_

def merge_chunk(final_response: dict, delta: dict) -> None:
    delta.pop("role", None)
    merge_fields(final_response, delta)

    tool_calls = delta.get("tool_calls")
    if tool_calls:
        for tool_call in tool_calls:
            index = tool_call.pop("index")
            # type_ = tool_call.pop("type", None)
            final_tool_call = final_response["tool_calls"][index]
            merge_fields(final_tool_call, tool_call)
            # if type_ and not final_tool_call.get("type"):
            #     # type = "function" is always returned by LiteLLM in the delta. Bug?
            #     # This ensures that the type is only set once.
            #     final_tool_call["type"] = type_

completion_params = {
    "stream": True,
    "caching": True,
    "temperature": 0.0,
    "model": "gpt-4.1",
    "tools": [
        {
            'type': 'function', 
            'function': {
                'name': 'calculator', 
                'description': 
                'Calculate the result of the expression.', 
                'parameters': {
                    'type': 'object', 
                    'properties': {
                        'expression': {'type': 'string'}
                    }, 
                    'required': ['expression']
                }
            }
        }
    ],
    "messages": [
        {
            "role": "system",
            "content": dedent(
                """\
                # RULES YOU MUST FOLLOW
                - You always use the calculator to answer questions. 
                - You always generate two calculator tool calls in your response.
                - You answer with a comma separated list of numbers.
                """
            ),
        },
        {
            "role": "user",
            "content": "What is 2 + 2 and what is 4 * 4?"
        },
    ]
}

message_template = {
    "content": "",
    "role": "assistant",
    "function_call": None,
    "tool_calls": defaultdict(
        lambda: {
            "function": {"arguments": "", "name": ""},
            "id": "",
            "type": "",
        }
    ),
}

async def run():
    completion = await litellm.acompletion(**completion_params)
    message = copy.deepcopy(message_template)
    async for chunk in completion:
        delta = chunk.choices[0].delta.model_dump()
        print("DELTA:")
        print(delta)
        print("---")
        delta.pop("role", None)
        merge_chunk(message, delta)

    result1 = str(message)
    print("RESULT1:")
    print(result1)
    print("---")

    print("================================================")

    completion = await litellm.acompletion(**completion_params)
    message = copy.deepcopy(message_template)
    async for chunk in completion:
        delta = chunk.choices[0].delta.model_dump()
        print("DELTA:")
        print(delta)
        print("---")
        delta.pop("role", None)
        merge_chunk(message, delta)

    result2 = str(message)
    print("RESULT2:")
    print(result2)
    print("---")
    assert result1 == result2

import asyncio
asyncio.run(run())

"""
{'content': '', 'role': 'assistant', 'function_call': None, 'tool_calls': defaultdict(<function <lambda> at 0x12a9f1c60>, {0: {'function': {'arguments': '{"expression": "2 + 2"}', 'name': 'calculator'}, 'id': 'call_o7S5j3ovMqFAQqJKk74t58mX', 'type': 'function'}, 1: {'function': {'arguments': '{"expression": "4 * 4"}', 'name': 'calculator'}, 'id': 'call_ev7jYDWHG7KOIrHch6PhDWCz', 'type': 'function'}})}
{'content': '', 'role': 'assistant', 'function_call': None, 'tool_calls': defaultdict(<function <lambda> at 0x12a9f1c60>, {0: {'function': {'arguments': '{"expression": "2 + 2"}', 'name': 'calculator'}, 'id': 'call_2kQY4KtxmThZFVT5C1fXYlGi', 'type': 'function'}, 1: {'function': {'arguments': '{"expression": "4 * 4"}', 'name': 'calculator'}, 'id': 'call_6eq6HnTs5ivrwfy2zVpV3sTj', 'type': 'function'}})}
{'content': '', 'role': 'assistant', 'function_call': None, 'tool_calls': defaultdict(<function <lambda> at 0x119121c60>, {0: {'function': {'arguments': '{"expression": "2 + 2"}', 'name': 'calculator'}, 'id': 'call_o7S5j3ovMqFAQqJKk74t58mX', 'type': 'function'}})}


{'content': '', 'role': 'assistant', 'function_call': None, 'tool_calls': defaultdict(<function <lambda> at 0x117506b60>, {0: {'function': {'arguments': '{"expression": "2 + 2"}', 'name': 'calculator'}, 'id': 'call_CJKTvRnUruMByt5ceQ2xVEbJ', 'type': 'function'}, 1: {'function': {'arguments': '{"expression": "4 * 4"}', 'name': 'calculator'}, 'id': 'call_a0iKQjJC2hbrbs7e3IEjXWMS', 'type': 'function'}})}
{'content': '', 'role': 'assistant', 'function_call': None, 'tool_calls': defaultdict(<function <lambda> at 0x117506b60>, {0: {'function': {'arguments': '{"expression": "2 + 2"}', 'name': 'calculator'}, 'id': 'call_HWKbzcVLXx4pkAoSGr6Ers2p', 'type': 'function'}, 1: {'function': {'arguments': '{"expression": "4 * 4"}', 'name': 'calculator'}, 'id': 'call_5bBE8QgrswsVk5S2OLDl4x3t', 'type': 'function'}})}
{'content': '', 'role': 'assistant', 'function_call': None, 'tool_calls': defaultdict(<function <lambda> at 0x107aeab60>, {0: {'function': {'arguments': '{"expression": "2 + 2"}', 'name': 'calculator'}, 'id': 'call_CJKTvRnUruMByt5ceQ2xVEbJ', 'type': 'function'}, 1: {'function': {'arguments': '{"expression": "4 * 4"}', 'name': 'calculator'}, 'id': 'call_a0iKQjJC2hbrbs7e3IEjXWMS', 'type': 'function'}})}
{'content': '', 'role': 'assistant', 'function_call': None, 'tool_calls': defaultdict(<function <lambda> at 0x123d12b60>, {0: {'function': {'arguments': '{"expression": "2 + 2"}', 'name': 'calculator'}, 'id': 'call_CJKTvRnUruMByt5ceQ2xVEbJ', 'type': 'function'}, 1: {'function': {'arguments': '{"expression": "4 * 4"}', 'name': 'calculator'}, 'id': 'call_a0iKQjJC2hbrbs7e3IEjXWMS', 'type': 'function'}})}
"""



import asyncio
from datetime import datetime
import logging
import random

import aiohttp
from markdownify import markdownify as md
from bs4 import BeautifulSoup
from slimagents import Agent, ToolResult
from dotenv import load_dotenv
from pydantic import BaseModel
from slimagents.repl import run_demo_loop

load_dotenv()

def calculator(expression: str) -> float:
    """Calculate the result of an expression. IMPORTANT: Always use this tool for calculations, never try to do it yourself."""
    print(f"--- Calculating {expression}")  
    ret = eval(expression)
    print("-->", str(ret), "\n")
    return ret

def python_evaluator(expression: str) -> str:
    """Evaluate a Python expression. Use this tool for calculations and other complex operations."""
    print(f"--- Evaluating {expression}")  
    ret = str(eval(expression))
    print("-->", str(ret), "\n")
    return ret

def show_on_screen(text: str) -> str:
    """Show text on the screen. This tool is used to display important messages. The returned value describes if the operation was successful or not."""
    print(f"--- Showing on screen: {text}")
    ret = random.choice(["success", "failure"])
    print("-->", ret, "\n")
    # return "Unable to show on screen. Screen is broken."
    return ret

def get_profit_and_loss_report(month: int, year: int) -> str:
    """Get the profit and loss report for a given month and year.
    Arguments:
        month (int): The month to get the profit and loss report for. 1 - 12.
        year (int): The year to get the profit and loss report for.
    """
    print(f"--- Getting profit and loss report for {month}/{year}")
    ret = """
    | Account               | Amount  |
    |-----------------------|---------|
    | Sales                 | 1000    |
    | Cost of goods sold    | 500     |
    | Other expenses        | 300     |
    """
    print("-->", ret, "\n")
    return ret

def get_weather(location: str, year: int, month: int, day: int) -> str:
    """Get the weather for a given location on a given date."""
    print(f"--- Getting weather for {location} on {year:04d}-{month:02d}-{day:02d}")
    temperature = random.randint(10, 30)
    weather = random.choice(["sunny", "cloudy", "rainy", "snowy"])
    ret = f"It's {weather} and {temperature} degrees."
    print("-->", ret, "\n")
    return ret

def get_current_date(today: str) -> str:
    """Get the current date. Always use "today" as the input."""
    # Gemini does not work with no argument tool calls...
    print("--- Getting current date")
    ret = datetime.now().strftime("%Y-%m-%d")
    print("-->", ret, "\n")
    return ret

def thought(description: str):
    """Describe your thoughts. This is a special tool that you will always use first."""
    print(f"\n### Thought: {description}")
    return "Thought registered"

def final_answer(answer: str):
    """Register your final answer. Never use this tool in combination with any calls to normal tools."""
    print(f"--- Final answer: {answer}")
    return ToolResult(value=answer, is_final_answer=True)

async def get_web_page_content_as_markdown(url: str) -> str:
    """Get the content of a URL formatted as markdown."""
    print(f"Getting content of {url}")
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            html = await response.text()
            soup = BeautifulSoup(html, 'html.parser')
            
            # Remove script and style elements
            for script in soup(["script", "style"]):
                script.decompose()
                
            # Get text and clean it up
            text = soup.get_text()
            # Break into lines and remove leading/trailing space
            lines = (line.strip() for line in text.splitlines())
            # Break multi-headlines into a line each
            chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
            # Drop blank lines
            text = ' '.join(chunk for chunk in chunks if chunk)

            print(f"Got content of {url}")
            
            return text
        
async def duckduckgo_search(query: str) -> list[tuple[str, str]]:
    """Search DuckDuckGo for a query and return the search results as list of (url, description) pairs."""
    base_url = f"https://duckduckgo.com/html/?q={query}"
    async with aiohttp.ClientSession() as session:
        async with session.get(base_url) as response:
            html = await response.text()
            soup = BeautifulSoup(html, 'html.parser')
            
            # Extract search results
            results = []
            for result in soup.find_all('div', class_='result'):
                title = result.find('a', class_='result__a')
                url = title.get('href') if title else None
                description = result.find('a', class_='result__snippet')
                
                if title and url and description:
                    results.append(("https://"+url, description.text.strip()))
            
            return results[:5]  # Return top 5 results


instructions = """
# MISSION
You are a helpful assistant who excels at planning and executing tasks by using tools.

# INSTRUCTIONS
- Always use the provided tools to solve the problem, if possible. In particular, always use the python_evaluator tool if it is possible to use it.
- If you cannot use the tools to solve the problem, then you will have to use your own knowledge to solve the problem.
- Never make up answers. If you don't know the answer, then you should say so.
- Never generate tool calls that depend on the results of other tool calls that have not been executed yet.
- If you generate several tool calls it is important that they are independent, i.e. that they can be executed in any order.

# IMPORTANT
- The user does not have access to your thought process and the results of the tool calls. Take this into account when generating the final answer.
"""

python_function_writer_instructions = """
# IDENTITY
You are a helpful assistant who excels at writing Python code.

# MISSION
You will be given a problem to solve and you will have to write a Python function that solves the problem.

# INSTRUCTIONS
- You always try to generalize the function as much as possible. Create suitable arguments for the function.
- You always try to make the function as simple as possible.
"""

class MyAgent(Agent):
    def _before_chat_completion(self) -> None:
        print("### New iteration\n")

    def python_evaluator(self, expression: str) -> str:
        function_writer_agent = Agent(
            instructions="You are a helpful assistant who excels at writing Python code.",
            # tools=[],
        )
        function_writer_agent.run_sync(expression)

agent = Agent(
    instructions=instructions,
    tools=[
        python_evaluator, 
        calculator, 
        # show_on_screen, 
        get_profit_and_loss_report, 
        get_weather, 
        get_current_date, 
        duckduckgo_search, 
        get_web_page_content_as_markdown,
    ],
    model="o3-mini"
)


query = "What is the temperature difference between Oslo and Stockholm? Show the result on the screen."
query = "How many R's are in the word 'STRAWBERRY'?"

# print("QUERY:", query, "\n")
# response = agent.run_sync(query)
# print("ANSWER:", response.value, "\n")

run_demo_loop(agent, log_level=logging.INFO)



# query = "What is 3 + 3 * 4 - 2 / 2? Show the result on the screen."
# response = await agent.run("What is 3 + 3 * 4 - 2 / 2? Show the result on the screen.")

# query = "What was the profit last month? Show the result on the screen."
# print("QUERY:", query, "\n")
# response = agent.run_sync(query)
# print("ANSWER:", response.value, "\n")

# query = "Calculate the average temperature difference between Oslo and Stockholm the last 4 days."
# print("QUERY:", query, "\n")
# response = agent.run_sync(query)
# print("ANSWER:", response.value, "\n")




import logging
from slimagents import Agent, run_demo_loop
from slimagents.core import Response, ToolResult

logging.basicConfig(level=logging.INFO)

def calculator(input: str) -> str:
    """
    Calculates mathematical expressions.

    Args:
        input(str): A mathematical expression to calculate.

    Returns:
        float: The result of the mathematical expression.
    """
    return ToolResult(value=eval(input), is_final_answer=True)

agent = Agent(model="gpt-4o-mini", tools=[calculator], instructions="You always use the calculator tool to calculate mathematical expressions.")
agent.logger.setLevel(logging.INFO)

run_demo_loop(agent)
# run_demo_loop(agent, stream=True)



import logging
import aiohttp
from slimagents import Agent
from dotenv import load_dotenv
import asyncio
from markdownify import markdownify as md
from bs4 import BeautifulSoup

load_dotenv()

def add(a: float, b: float) -> float:
    """Add two numbers."""
    return a + b

def subtract(a: float, b: float) -> float:
    """Subtract two numbers."""
    return a - b

def multiply(a: float, b: float) -> float:
    """Multiply two numbers."""
    return a * b

def divide(a: float, b: float) -> float:
    """Divide two numbers."""
    return a / b

async def get_url_(url: str, max_chars: int = None) -> str:
    """Get the content of a URL. Use max_chars to limit the response if needed."""
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            return await response.text()[:max_chars] if max_chars else await response.text()

async def get_url(url: str) -> str:
    """Get the content of a URL."""
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            html = await response.text()

            return (await response.text())[:10000]

async def get_web_page_content_as_text(url: str) -> str:
    """Get the content of a URL as text, using BeautifulSoup to extract readable text."""
    print(f"Getting content of {url}")
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            html = await response.text()
            soup = BeautifulSoup(html, 'html.parser')
            
            # Remove script and style elements
            for script in soup(["script", "style"]):
                script.decompose()
                
            # Get text and clean it up
            text = soup.get_text()
            # Break into lines and remove leading/trailing space
            lines = (line.strip() for line in text.splitlines())
            # Break multi-headlines into a line each
            chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
            # Drop blank lines
            text = ' '.join(chunk for chunk in chunks if chunk)

            print(f"Got content of {url}")
            
            return text
        
async def duckduckgo_search(query: str) -> list[tuple[str, str]]:
    """Search DuckDuckGo for a query and return the search results as list of (url, description) pairs."""
    base_url = f"https://duckduckgo.com/html/?q={query}"
    async with aiohttp.ClientSession() as session:
        async with session.get(base_url) as response:
            html = await response.text()
            soup = BeautifulSoup(html, 'html.parser')
            
            # Extract search results
            results = []
            for result in soup.find_all('div', class_='result'):
                title = result.find('a', class_='result__a')
                url = title.get('href') if title else None
                description = result.find('a', class_='result__snippet')
                
                if title and url and description:
                    results.append(("https://"+url, description.text.strip()))
            
            return results[:5]  # Return top 5 results

def register_thoughts(thoughts: str):
    """Register your thoughts. Always use this tool when you are using another tool."""
    return "Thoughts registered"
        

agent = Agent(
    name="Agent",
    instructions="You are a helpful agent. You will always use the tools to solve the problem, if possible.",
    tools=[add, subtract, multiply, divide],
)

agent.logger.setLevel(logging.DEBUG)

#     agent = Agent(
#         name="Agent",
#         instructions="""You are a web research agent. You will always answer the question based on information found online using the provided tools.
# You will first analyze the question and determine what tools you need to use to answer the question. You will then use the register_thoughts tool to register your thoughts that describe what you are doing and why you are using the tool.
# You will keep calling tools until you have answered the question completely or you have no more tools to use.
# """,
#         functions=[register_thoughts, add, subtract, multiply, divide, get_web_page_content_as_text, duckduckgo_search],
#     )

# messages = [{"role": "user", "content": "Hva er hovedoppslaget på https://www.nrk.no, https://www.vg.no, https://www.aftenposten.no?"}]
# messages = [{"role": "user", "content": "Aside from the Apple Remote, what other device can control the program Apple Remote was originally designed to interact with?"}]
response = agent.run_sync("What is 3 + 3 * 4 - 2 / 2?")
print(response.value)

async def test():
    results = await duckduckgo_search("How fast can a human run?")
    print(results)

asyncio.run(test())


from slimagents import Agent

def python_evaluator(expression: str) -> str:
    """Evaluate a Python expression. Always use this tool for calculations and other complex operations."""
    print(f"--- Evaluating {expression}")
    # Obviously not secure, but for the sake of this example we'll just eval the expression.
    ret = str(eval(expression))
    print(f"--> {ret}")
    return ret

agent = Agent(
    instructions="You are a helpful assistant. When given a task you always try to solve it by using tools, never rely on your own knowledge.",
    tools=[python_evaluator],
)

prompt = "How many R's are in the word 'STRAWBERRY'?"
print(f"User: {prompt}")
response = agent.run_sync(prompt)
print(f"Agent: {response.value}")
print(response.memory_delta)


import base64

from openai import Client

filename = "temp/Get_Started_With_Smallpdf.pdf"
with open(filename, "rb") as f:
    file_content = f.read()
    base64_content = base64.b64encode(file_content).decode('utf-8')
    file_data = f"data:application/pdf;base64,{base64_content}"
    file = {
        "type": "file",
        "file": {
            "filename": filename,
            "file_data": file_data,
        },
    }
    text = "Extract the text from the PDF file in markdown format. Don't include any other text or formatting."
    openai_client = Client()
    response = openai_client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[{"role": "user", "content": [file, {"type": "text", "text": text}]}],
    )
    print(response.choices[0].message.content)

    # file_content = f.read()
    # base64_content = base64.b64encode(file_content).decode('utf-8')
    # file_data = f"data:application/pdf;base64,{base64_content}"
    # file = {
    #     "type": "file",
    #     "file": {
    #         "filename": "test.pdf",
    #         "file_data": file_data,
    #     },
    # }


    # from slimagents import Agent
    # agent = Agent(model="gpt-4o-mini")

    # result = agent.run_sync(file, "Extract the text from the PDF file.")
    # print(result.value)
    
    # import litellm
    # response = litellm.completion(
    #     model="gpt-4o-mini",
    #     messages=[{"role": "user", "content": [file, {"type": "text", "text": text}]}],
    # )
    # print(response.choices[0].message.content)
    



# !pip install python-weather

from slimagents.core import Agent
import python_weather

class WeatherAgent(Agent):
    def __init__(self):
        super().__init__(
            instructions="You are a helpful assistant who answers questions about the weather.",
            tools=[self.get_temperature],
        )

    async def get_temperature(self, location: str) -> float:
        """Get the current temperature in a given location, in degrees Celsius."""
        async with python_weather.Client(unit=python_weather.METRIC) as client:
            print(f"--- Getting temperature for {location}")
            weather = await client.get(location)
            print(f"--> Temperature in {location}: {weather.temperature}")
            return weather.temperature

agent = WeatherAgent()
prompt = "What is the temperature difference between London and Paris?"
print(f"User: {prompt}")
response = agent.run_sync(prompt)
print(f"Agent: {response.value}")



import asyncio
from datetime import datetime
import logging
import random
from slimagents import Agent, ToolResult
from dotenv import load_dotenv
from pydantic import BaseModel

load_dotenv()

def calculator(expression: str) -> float:
    """Calculate the result of an expression. IMPORTANT: Always use this tool for calculations, never try to do it yourself."""
    print(f"--- Calculating {expression}")  
    ret = eval(expression)
    print("-->", str(ret))
    return ret

def show_on_screen(text: str):
    """Show text on the screen."""
    print(f"--- Showing on screen: {text}")
    ret = "The message was shown on the screen."
    print("-->", ret)
    return ret

def get_profit_and_loss_report(month: int, year: int) -> str:
    """Get the profit and loss report for a given month and year.
    Arguments:
        month (int): The month to get the profit and loss report for. 1 - 12.
        year (int): The year to get the profit and loss report for.
    """
    print(f"--- Getting profit and loss report for {month}/{year}")
    ret = """
    | Account               | Amount  |
    |-----------------------|---------|
    | Sales                 | 1000    |
    | Cost of goods sold    | 500     |
    | Other expenses        | 300     |
    """
    print("-->", ret)
    return ret

def get_weather(location: str) -> str:
    """Get the weather for a given location."""
    print(f"--- Getting weather for {location}")
    temperature = random.randint(10, 30)
    weather = random.choice(["sunny", "cloudy", "rainy", "snowy"])
    ret = f"It's {weather} and {temperature} degrees."
    print("-->", ret)
    return ret

def get_current_date(today: str) -> str:
    """Get the current date. Always use "today" as the input."""
    print("--- Getting current date")
    ret = datetime.now().strftime("%Y-%m-%d")
    print("-->", ret)
    return ret

def thought(description: str):
    """Describe your thoughts. This is a special tool that you will always use first."""
    print(f"\n### Thought: {description}")
    return "Thought registered"

def final_answer(answer: str):
    """Register your final answer. Never use this tool in combination with any calls to normal tools."""
    print(f"--- Final answer: {answer}")
    return ToolResult(value=answer, is_final_answer=True)



instructions = """
# INSTRUCTIONS
You are a helpful assistant. You have access to a set of tools, and you will always use the tools to generate the answer, if possible. 

# TOOLS

Two of the tools are special:
- "thought": Describe your thought.
- "final_answer": Provide the final answer to the user's question.
These tools will be referred to as "special tools".

The other tools are tools that can actually be used to solve the problem. These tools will be referred to as "normal tools".

# RESPONSE PROTOCOL

## WITH NORMAL TOOL CALLS
Use this protocol when you will use normal tools to solve the problem.
- Always start by generating a tool call to the "thought" tool. Describe which tool calls you will make and why.
- Then generate one or more tool calls to actually solve the problem, if you have not reached a solution yet.

IMPORTANT
- In this scenario, you will never generate a tool call to the "final_answer" tool.
- Only generate tool calls that are independent, i.e. they do not depend on the results of other tool calls that have not been executed yet.

## WITH SPECIAL TOOL CALLS
Use this protocol when you have reached a solution based on normal tool calls, or if you can answer the user's question directly.
- Always start by generating a tool call to the "thought" tool. Describe why you think you have reached a solution and how you will formulate the final answer.
- Then generate a tool call to the "final_answer" tool containing the final answer.

IMPORTANT
- In this scenario, you will never generate a tool call to any normal tool.
- The user does not have access to your thought process and the results of the tool calls. Take this into account when generating the final answer. Make sure to include all relevant information.
"""

async def main():
    agent = Agent(
        instructions=instructions,
        # tools=[thought, final_answer, add, subtract, multiply, divide],
        # tools=[thought, final_answer, calculator, show_on_screen, get_profit_and_loss_report, get_weather],
        tools=[thought, final_answer, calculator, show_on_screen, get_profit_and_loss_report, get_weather, get_current_date],
        tool_choice="required",
        # model="gemini/gemini-1.5-flash-002",
        # model="gemini/gemini-2.0-flash-exp",
        # parallel_tool_calls=None,
        # response_format=MyResult,
        # extra_llm_params={"drop_params": True},
    )

    agent.logger.setLevel(logging.DEBUG)

    # response = await agent.run("What is 3 + 3 * 4 - 2 / 2? Show the result on the screen.")
    # response = await agent.run("What was the profit last month? Show the result on the screen.")
    response = await agent.run("What is the temperature difference betweeen Oslo and Stockholm? Show the result on the screen.")
    print(response.value)


if __name__ == "__main__":
    asyncio.run(main())



import litellm
from litellm.caching.caching import Cache

litellm.cache = Cache()

completion_params = {
    "stream": True,
    "caching": True,
    "temperature": 0.0,
    "model": "gpt-4.1",
    "messages": [
        {
            "role": "user",
            "content": "Count to 10"
        },
    ]
}

completion = litellm.completion(**completion_params)
deltas1 = []
for chunk in completion:
    delta = chunk.choices[0].delta.model_dump()
    deltas1.append(str(delta))
    print(f"DELTA: {delta}")

print("================================================")

completion = litellm.completion(**completion_params)
deltas2 = []
for chunk in completion:
    delta = chunk.choices[0].delta.model_dump()
    deltas2.append(str(delta))
    print(f"DELTA: {delta}")

assert deltas1 == deltas2



from slimagents import Agent
from slimagents.repl import run_demo_loop

class StrictAgent(Agent):
    def __init__(self, max_responses: int):
        super().__init__(
            tools=[self.update_responses_left],
        )
        self._answers_left = max_responses + 1

    @property
    def instructions(self) -> str:
        if self._answers_left > 0:
            return f"""
You are a helpful assistant. 
You currently have {self._answers_left} responses left.
ALWAYS call the `update_responses_left` tool before you respond."""
        else:
            return "You always answer 'I can't answer that.'."

    def update_responses_left(self):
        """IMPORTANT! You ALWAYS call this tool before you respond, no matter what the user says."""
        self._answers_left -= 1
        return "Good! You may now answer the question."

agent = StrictAgent(2)
run_demo_loop(agent)



from slimagents import Agent
from pydantic import BaseModel, Optional
from typing import List


class OrderItem(BaseModel):
    Beskrivelse: str
    Antall: Optional[float] = None
    Enhetspris: Optional[float] = None
    Totalpris_ink_MVA: Optional[float] = None

class OrderConfirmation(BaseModel):
    Ordrenummer: Optional[str]
    Leveransedato: Optional[str]
    Ankomsttid: Optional[str]
    Serveringstid: Optional[str]
    Kunde: Optional[str]
    Adresse: Optional[str]
    Kundens_referanse: Optional[str]
    Kunde_tlf: Optional[str]
    Kontaktperson: Optional[str]
    Postnr: Optional[str]
    Poststed: Optional[str]
    Order_Items: List[OrderItem] = []
    Total: Optional[float] = None


pdf_converter = Agent(
    instructions="Your task is to convert PDF files to Markdown.",
    model="gemini/gemini-2.0-flash",
)

# response = agent.run_sync("Who are you?")
# print(response.value)

with open("./temp/Enchiladas med salat.pdf", "rb") as pdf_file:
    response = pdf_converter.run_sync(pdf_file)
    print(response.value)



# def foo(a, b, c=3, /, *, d):
#     print(a, b, c)

# foo(1, 2, 3)


from slimagents.core import Agent, ToolResult


def calculator(expression: str) -> float:
    """
    Calculate the result of the expression.
    """
    value = eval(expression)
    return ToolResult(value=value)

calc_agent = Agent(
    name="Calculator",
    instructions="You are not good at math, but you have a calculator.",
    tools=[calculator],
    model="gpt-4o-mini",
)

memory = []

calc_agent.apply("What is 2 + 2?", memory=memory)

print(memory)



import asyncio
from datetime import datetime
import logging
import random
from slimagents import Agent, ToolResult
from dotenv import load_dotenv
from pydantic import BaseModel

load_dotenv()

def calculator(expression: str) -> float:
    """Calculate the result of an expression. IMPORTANT: Always use this tool for calculations, never try to do it yourself."""
    print(f"--- Calculating {expression}")  
    ret = eval(expression)
    print("-->", str(ret), "\n")
    return ret

def show_on_screen(text: str) -> str:
    """Show text on the screen. The returned value describes if the task was successful or not."""
    print(f"--- Showing on screen: {text}")
    ret = random.choice(["success", "failure"])
    print("-->", ret, "\n")
    return ret

def get_profit_and_loss_report(month: int, year: int) -> str:
    """Get the profit and loss report for a given month and year.
    Arguments:
        month (int): The month to get the profit and loss report for. 1 - 12.
        year (int): The year to get the profit and loss report for.
    """
    print(f"--- Getting profit and loss report for {month}/{year}")
    ret = """
    | Account               | Amount  |
    |-----------------------|---------|
    | Sales                 | 1000    |
    | Cost of goods sold    | 500     |
    | Other expenses        | 300     |
    """
    print("-->", ret, "\n")
    return ret

def get_weather(location: str, year: int, month: int, day: int) -> str:
    """Get the weather for a given location on a given date."""
    print(f"--- Getting weather for {location} on {year:04d}-{month:02d}-{day:02d}")
    temperature = random.randint(10, 30)
    weather = random.choice(["sunny", "cloudy", "rainy", "snowy"])
    ret = f"It's {weather} and {temperature} degrees."
    print("-->", ret, "\n")
    return ret

def get_current_date(today: str) -> str:
    """Get the current date. Always use "today" as the input."""
    # Gemini does not work with no argument tool calls...
    print("--- Getting current date")
    ret = datetime.now().strftime("%Y-%m-%d")
    print("-->", ret, "\n")
    return ret

def thought(description: str):
    """Describe your thoughts. This is a special tool that you will always use first."""
    print(f"\n### Thought: {description}")
    return "Thought registered"

def final_answer(answer: str):
    """Register your final answer. Never use this tool in combination with any calls to normal tools."""
    print(f"--- Final answer: {answer}")
    return ToolResult(value=answer, is_final_answer=True)



instructions = """
# MISSION
You are a helpful assistant who excels at planning and executing tasks by using tools.

# INSTRUCTIONS
- Always use the provided tools to solve the problem, if possible. 
- If you cannot use the tools to solve the problem, then you will have to use your own knowledge to solve the problem.
- Never make up answers. If you don't know the answer, then you should say so.
- Never generate tool calls that depend on the results of other tool calls that have not been executed yet.
- If you generate several tool calls it is important that they are independent, i.e. they can be executed in any order.

# IMPORTANT
- The user does not have access to your thought process and the results of the tool calls. Take this into account when generating the final answer.
"""

class MyAgent(Agent):
    def _before_chat_completion(self) -> None:
        print("### New iteration\n")

agent = MyAgent(
    instructions=instructions,
    # tools=[thought, final_answer, add, subtract, multiply, divide],
    # tools=[thought, final_answer, calculator, show_on_screen, get_profit_and_loss_report, get_weather],
    tools=[calculator, show_on_screen, get_profit_and_loss_report, get_weather, get_current_date],
    # tool_choice="required",
    # model="gemini/gemini-1.5-flash-002",
    # model="gemini/gemini-2.0-flash-exp",
    parallel_tool_calls=None,
    # response_format=MyResult,
    # extra_llm_params={"drop_params": True},
)


agent.logger.setLevel(logging.DEBUG)

# query = "What is 3 + 3 * 4 - 2 / 2? Show the result on the screen."
# response = await agent.run("What is 3 + 3 * 4 - 2 / 2? Show the result on the screen.")

# query = "What was the profit last month? Show the result on the screen."
# print("QUERY:", query, "\n")
# response = agent.run_sync(query)
# print("ANSWER:", response.value, "\n")

query = "What is the temperature difference betweeen Oslo and Stockholm? Show the result on the screen."
print("QUERY:", query, "\n")
response = agent.run_sync(query)
print("ANSWER:", response.value, "\n")

# query = "Calculate the average temperature difference between Oslo and Stockholm the last 4 days."
# print("QUERY:", query, "\n")
# response = agent.run_sync(query)
# print("ANSWER:", response.value, "\n")




import litellm
from litellm.caching.caching import Cache

# litellm.cache = Cache(type="disk", disk_cache_dir="./test_cache")
litellm.cache = Cache()

completion_params = {
    "stream": True,
    "caching": True,
    "temperature": 0.0,
    "model": "gpt-4.1",
    "tools": [
        {
            'type': 'function', 
            'function': {
                'name': 'calculator', 
                'description': 
                'Calculate the result of the expression.', 
                'parameters': {
                    'type': 'object', 
                    'properties': {
                        'expression': {'type': 'string'}
                    }, 
                    'required': ['expression']
                }
            }
        }
    ],
    "messages": [
        {
            "role": "user",
            "content": "Count to 10"
        },
    ]
    # "messages": [
    #     {
    #         "role": "system",
    #         "content": "You always use the calculator to answer questions."
    #     },
    #     {
    #         "role": "user",
    #         "content": "What is 2 + 2?"
    #     },
    # ]
}

async def run():
    completion = await litellm.acompletion(**completion_params)
    deltas1 = []
    async for chunk in completion:
        delta = chunk.choices[0].delta.model_dump()
        deltas1.append(str(delta))
        print(f"DELTA: {delta}")
        print("---")

    print("================================================")

    completion = await litellm.acompletion(**completion_params)
    deltas2 = []
    async for chunk in completion:
        delta = chunk.choices[0].delta.model_dump()
        deltas2.append(str(delta))
        print(f"DELTA: {delta}")
        print("---")

    assert deltas1 == deltas2

def run_sync():
    completion = litellm.completion(**completion_params)
    deltas1 = []
    for chunk in completion:
        delta = chunk.choices[0].delta.model_dump()
        deltas1.append(str(delta))
        print(f"DELTA: {delta}")
        print("---")

    print("================================================")

    completion = litellm.completion(**completion_params)
    deltas2 = []
    for chunk in completion:
        delta = chunk.choices[0].delta.model_dump()
        deltas2.append(str(delta))
        print(f"DELTA: {delta}")
        print("---")

    assert deltas1 == deltas2

run_sync()

# import asyncio
# asyncio.run(run())



import copy
import inspect
import re
from typing import Any, Generic, Optional, TypeVar, get_args, get_origin
import jsonref
from pydantic import BaseModel
import mimetypes
from slimagents.config import logger

_has_magic = False
try:
    from magic import Magic
    _has_magic = True
except ImportError as e:
    logger.warning(f"python-magic is not properly installed, content based MIME type detection will be disabled. Cause: {e}")
    pass


def merge_fields(target, source):
    for key, value in source.items():
        if isinstance(value, str):
            target[key] += value
        elif value is not None and isinstance(value, dict):
            merge_fields(target[key], value)


def merge_chunk(final_response: dict, delta: dict) -> None:
    delta.pop("role", None)
    merge_fields(final_response, delta)

    tool_calls = delta.get("tool_calls")
    if tool_calls:
        for tool_call in tool_calls:
            index = tool_call.pop("index")
            type_ = tool_call.pop("type", None)
            final_tool_call = final_response["tool_calls"][index]
            merge_fields(final_tool_call, tool_call)
            if type_ and not final_tool_call.get("type"):
                # type = "function" is always returned by LiteLLM in the delta. Bug?
                # This ensures that the type is only set once.
                final_tool_call["type"] = type_


def function_to_json(func) -> dict:
    """
    Converts a Python function into a JSON-serializable dictionary
    that describes the function's signature, including its name,
    description, and parameters.

    Args:
        func: The function to be converted.

    Returns:
        A dictionary representing the function's signature in JSON format.
    """
    type_map = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
        type(None): "null",
    }

    try:
        signature = inspect.signature(func)
    except ValueError as e:
        raise ValueError(
            f"Failed to get signature for function {func.__name__}: {str(e)}"
        )

    parameters = {}
    for param in signature.parameters.values():
        try:
            param_type = type_map.get(param.annotation, "string")
        except KeyError as e:
            raise KeyError(
                f"Unknown type annotation {param.annotation} for parameter {param.name}: {str(e)}"
            )
        parameters[param.name] = {"type": param_type}

    required = [
        param.name
        for param in signature.parameters.values()
        if param.default == inspect._empty
    ]

    return {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": func.__doc__ or "",
            "parameters": {
                "type": "object",
                "properties": parameters,
                "required": required,
            },
        },
    }


def fix_schema(schema):
    if "title" in schema:
        # Title is not very useful, and also not supported by Gemini
        del schema["title"]
    if "default" in schema:
        # Default is not supported by OpenI or Gemini
        del schema["default"]
    _type = schema.get("type")
    if _type == "array":
        fix_schema(schema["items"])
    elif _type == "object":
        # OpenAI requires additionalProperties to be false
        schema["additionalProperties"] = False
        required = []
        for propery, property_type in schema["properties"].items():
            fix_schema(property_type)
            # OpenAI requires all properties to be required
            required.append(propery)
        schema["required"] = required
    return schema


def flatten_schema(schema):
    if "$defs" in schema:
        schema = jsonref.replace_refs(schema, lazy_load=False)
        del schema["$defs"]
        # Convert non-JSON-serializable types to serializable types
        schema = copy.deepcopy(schema)
    return schema


T = TypeVar('T')

class PrimitiveResult(BaseModel):
    result: Any

class IntResult(PrimitiveResult):
    result: int

class FloatResult(PrimitiveResult):
    result: float

class BoolResult(PrimitiveResult):
    result: bool

class ListResult(PrimitiveResult, Generic[T]):
    result: list[T]

TYPE_MAP = {
    int: IntResult,
    float: FloatResult,
    bool: BoolResult,
    list: ListResult,
}


def get_pydantic_type(t: type) -> type:
    _t = get_origin(t) or t
    ret = TYPE_MAP.get(_t, t)
    if ret is ListResult:
        t_args = get_args(t)
        if len(t_args) == 0:
            return ret[Any]
        elif len(t_args) == 1:
            return ret[t_args[0]]
        else:
            raise ValueError(f"Unsupported list type: {t}")
    return ret


JSON_MODE = {
    "type": "json_object",
}

def type_to_response_format(type_: Optional[type]) -> dict:
    if type_ is None:
        return None
    elif isinstance(type_, dict):
        return type_
    elif type_ is dict:
        return JSON_MODE
    elif issubclass(type_, BaseModel):
        schema = type_.model_json_schema(mode="serialization")
        # LLMs typically don't support $defs, so we need to remove them
        schema = flatten_schema(schema)
        schema = fix_schema(schema)

        ret = {
            "type": "json_schema",
            "json_schema": {
                "name": re.sub(r'[^a-zA-Z0-9-_]', '_', type_.__name__),
                "schema": schema,
                "strict": True,
            },
        }
        return ret
    else:
        raise ValueError(f"Unsupported type for response_format: {type_}")
    

def get_mime_type_from_file_like_object(file_like_object:str, filename:str = None):
    """
    Determines the MIME type of a file-like object by examining its name (if available)
    or its content as a fallback.
    
    Args:
        file_like_object: A file-like object with read and seek methods
        filename: The name of the file to get the MIME type from
    Returns:
        str: The detected MIME type of the content, or 'application/octet-stream' if detection fails
    """
    # First try to get MIME type from filename if available
    if filename:
        mime_type, _ = mimetypes.guess_type(filename)
        if mime_type:
            return mime_type
    
    # Try content-based detection if python-magic is available
    if _has_magic:
        current_pos = file_like_object.tell()
        try:
            # Read the first 2048 bytes for MIME detection
            content = file_like_object.read(2048)
            return get_mime_type_from_content(content)
        finally:
            # Restore original position
            file_like_object.seek(current_pos)
    return 'application/octet-stream'


def get_mime_type_from_content(content: bytes) -> str:
    """
    Determines the MIME type of a bytes object by examining its content.

    Args:
        content: A bytes object - usually the first 2048 bytes of a file

    Returns:
        str: The detected MIME type of the content, or 'application/octet-stream' if detection fails
    """
    # Try content-based detection if python-magic is available
    if _has_magic:
        try:
            # Create a Magic instance for MIME type detection
            mime = Magic(mime=True)
            
            # Detect MIME type
            if isinstance(content, str):
                mime_type = mime.from_buffer(content.encode('utf-8'))
            else:
                mime_type = mime.from_buffer(content)
                
            return mime_type
        except Exception:
            pass
            
    # If python-magic is not available or fails, return a default mime type
    return 'application/octet-stream'



# Standard library imports
import base64
from dataclasses import dataclass, field
from enum import Enum
import json
from collections import defaultdict
import random
import string
import time
from typing import AsyncGenerator, Callable, Optional, Union, Coroutine, Any
import inspect
import asyncio
import logging

# Package/library imports
from litellm import acompletion
from litellm.types.completion import ChatCompletionMessageToolCallParam, Function
from pydantic import AnyUrl, BaseModel

# Local imports
from .util import PrimitiveResult, function_to_json, get_mime_type_from_content, get_mime_type_from_file_like_object, get_pydantic_type, merge_chunk, type_to_response_format
import slimagents.config as config

# Types
AgentFunction = Callable[..., Union[str, "Agent", dict, Coroutine[Any, Any, Union[str, "Agent", dict]]]]

@dataclass
class Response():
    value: Any
    memory_delta: list
    agent: "Agent"

@dataclass
class ToolResult():
    """
    Encapsulates the possible return values for an agent tool call.

    Attributes:
        value (str): The result value as a string.
        agent (Agent): The agent instance, if applicable.
        is_final_answer (bool): Whether to exit the current agent and return the result as the final answer. Defaults to False.
        handoff (bool): Only used if an agent is provided. If true, the control of the conversation is transferred to the
                       provided agent. If false, the inputs are processed by the provided agent and the result is returned
                       as the tool call result.
    """
    value: str = ""
    agent: Optional["Agent"] = None
    is_final_answer: bool = False
    handoff: bool = False

@dataclass
class HandleToolCallResult():
    messages: list
    agent: Optional["Agent"] = None
    filtered_tool_calls: list[ChatCompletionMessageToolCallParam] = field(default_factory=list)
    result: Optional[ToolResult] = None

class Delimiter(Enum):
    ASSISTANT_START = "assistant_start"
    ASSISTANT_END = "assistant_end"
    TOOL_CALL = "tool_call"

@dataclass
class MessageDelimiter():
    """
    A delimiter for the message stream.
    """
    delimiter: Delimiter
    message: dict

# Agent class

DEFAULT_MODEL = "gpt-4.1"

class Agent:

    logger = config.agent_logger.getChild("Agent")

    def __init__(
            self, 
            name: Optional[str] = None, 
            model: Optional[str] = None,
            instructions: Optional[Union[str, Callable[[], str]]] = None, 
            memory: Optional[list[dict]] = None,
            tools: Optional[list[AgentFunction]] = None, 
            tool_choice: Optional[Union[str, dict]] = None, 
            parallel_tool_calls: Optional[bool] = None, 
            response_format: Optional[Union[dict, type[BaseModel]]] = None,
            temperature: Optional[float] = None,
            logger: Optional[logging.Logger] = None,
            **lite_llm_args
    ):
        self._name = name or self.__class__.__name__
        self._model = model or DEFAULT_MODEL
        self._instructions = instructions
        self._memory = memory or []
        self._tools = tools or []
        self._tool_choice = tool_choice
        self._parallel_tool_calls = parallel_tool_calls
        self._response_format = get_pydantic_type(response_format)
        self._temperature = temperature
        self._lite_llm_args = lite_llm_args

        # Set up logging
        if logger:
            self.logger = logger
        elif config.separate_agent_logger and self.__class__ != Agent:
            logger_name = f"{self.__class__.__module__}.{self.__class__.__name__}"
            self.logger = config.agent_logger.getChild(logger_name)
        else:
            # Use the class level logger
            pass

        # Cache related
        self.__tools = None
        self.__json_tools = None
        self.__json_response_format = None
        self.__all_chat_completion_params = None
        
    @property 
    def name(self):
        return self._name
    @name.setter
    def name(self, value):
        self._name = value
    
    @property
    def model(self):
        return self._model
    @model.setter
    def model(self, value):
        if value != self._model:
            self.__all_chat_completion_params = None
            self._model = value

    @property
    def instructions(self):
        return self._instructions
    @instructions.setter
    def instructions(self, value):
        if value != self._instructions:
            self.__all_chat_completion_params = None
            self._instructions = value

    @property
    def memory(self):
        """
        The "default" memory of the agent that will always be included for each chat completion.
        """
        return self._memory
    @memory.setter
    def memory(self, value):
        self._memory = value

    @property
    def tools(self):
        return self._tools
    @tools.setter
    def tools(self, value):
        if value != self._tools:
            self.__all_chat_completion_params = None
            self.__json_tools = None
            self._tools = value 

    @property
    def tool_choice(self):
        return self._tool_choice
    @tool_choice.setter
    def tool_choice(self, value):
        if value != self._tool_choice:
            self.__all_chat_completion_params = None
            self._tool_choice = value

    @property
    def parallel_tool_calls(self):
        return self._parallel_tool_calls
    @parallel_tool_calls.setter
    def parallel_tool_calls(self, value):
        if value != self._parallel_tool_calls:
            self.__all_chat_completion_params = None
            self._parallel_tool_calls = value

    @property
    def response_format(self):
        return self._response_format
    @response_format.setter
    def response_format(self, value):
        if value != self._response_format:
            self.__all_chat_completion_params = None
            self.__json_response_format = None
            self._response_format = get_pydantic_type(value)

    @property
    def temperature(self):
        return self._temperature
    @temperature.setter
    def temperature(self, value):
        if value != self._temperature:
            self.__all_chat_completion_params = None
            self._temperature = value

    @property
    def lite_llm_args(self):
        return self._lite_llm_args
    @lite_llm_args.setter
    def lite_llm_args(self, value):
        if value != self._lite_llm_args:
            self.__all_chat_completion_params = None
            self._lite_llm_args = value


    def __get_all_chat_completion_params(self):
        if self.__all_chat_completion_params is not None:
            if self.__tools == self.tools:
                # It's safe to return the cached params
                return self.__all_chat_completion_params
            else:
                # Tools list has changed from the "outside". Make sure to update the cache afterwards.
                self.__tools = None
                self.__json_tools = None
        if self.__tools != self.tools:
            # Tools list has changed from the "outside"
            self.__tools = self.tools
            if self.tools:
                self.__json_tools = [function_to_json(f) for f in self.tools]
            else:
                self.__json_tools = None
        if self.__json_response_format is None:
            # Response format is updated, so we need to update the cached JSON response format
            self.__json_response_format = type_to_response_format(self.response_format)
        params = {}
        if self._lite_llm_args:
            params.update(self._lite_llm_args)
        params.update({
            "model": self.model,
            "temperature": self.temperature,
        })
        if self.__json_tools:
            params.update({
                "tools": self.__json_tools,
                "tool_choice": self.tool_choice,
                "parallel_tool_calls": self.parallel_tool_calls,
            })
        if self.__json_response_format:
            params["response_format"] = self.__json_response_format
        return params


    async def _get_chat_completion(self, run_id: str, turns: int, memory: list[dict], memory_delta: list[dict], stream: bool = False, caching: bool = False):
        if self.instructions:
            messages = [{"role": "system", "content": self.instructions}]
        else:
            messages = []
        # self.memory is the "default" memory that will always be included for each chat completion
        messages.extend(self.memory)
        # Add the memory added by the user
        messages.extend(memory)
        # Add the memory added by the agent during the current call
        messages.extend(memory_delta)
        if self.logger.getEffectiveLevel() <= logging.DEBUG:
            self.logger.debug("Run %s-%d: Getting chat completion for: %s", run_id, turns, messages)
        else:
            self.logger.info("Run %s-%d: Getting chat completion for %d messages", run_id, turns, len(messages))

        create_params = self.__get_all_chat_completion_params().copy()
        create_params["messages"] = messages
        create_params["stream"] = stream
        create_params["caching"] = caching
        return await acompletion(**create_params)


    async def _handle_function_result(self, run_id: str, result, memory: list[dict], memory_delta: list[dict], caching: bool) -> ToolResult:
        if isinstance(result, ToolResult):
            if result.agent and not result.handoff:
                response = await result.agent._run(run_id, memory=memory.copy(), memory_delta=memory_delta.copy(), caching=caching)
                result.value = response.value
                result.agent = None
                return result
            else:
                return result
        elif isinstance(result, Agent):
            return ToolResult(
                value=json.dumps({"assistant": result.name}),
                agent=result,
            )
        else:
            try:
                return ToolResult(value=str(result))
            except Exception as e:
                error_message = "Failed to cast response to string: %s. Make sure agent functions return a string, Result object, or coroutine. Error: %s"
                self.logger.error(error_message, result, str(e))
                raise TypeError(error_message % (result, str(e)))
            

    def _get_value(self, content: str):
        if self.response_format:
            if self.response_format is dict or isinstance(self.response_format, dict):
                return json.loads(content)
            elif issubclass(self.response_format, BaseModel):
                ret = self.response_format.model_validate_json(content)
                if isinstance(ret, PrimitiveResult):
                    return ret.result
                else:
                    return ret
            else:
                raise ValueError(f"Unsupported response_format: {self.response_format}")
        else:
            return content


    def _update_partial_response(
            self, 
            partial_response: HandleToolCallResult, 
            tool_call: ChatCompletionMessageToolCallParam, 
            result: ToolResult
    ) -> None:
        partial_response.filtered_tool_calls.append(tool_call)
        partial_response.messages.append(
            {
                "role": "tool",
                "tool_call_id": tool_call["id"],
                "tool_name": tool_call["function"]["name"],
                "content": str(result.value),
            }
        )
        if result.agent:
            partial_response.agent = result.agent
        if result.is_final_answer:
            partial_response.result = result


    def _before_chat_completion(self) -> None:
        pass

    async def _handle_tool_calls(
            self,
            run_id: str,
            turn: int,
            tool_calls: list[ChatCompletionMessageToolCallParam],
            memory: list[dict],
            memory_delta: list[dict],
            caching: bool,
    ) -> HandleToolCallResult:
        function_map = {f.__name__: f for f in self.tools}
        partial_response = HandleToolCallResult(messages=[], agent=None, filtered_tool_calls=[])

        async_tasks = []
        for tool_call in tool_calls:
            function = tool_call["function"]
            name = function["name"]
            tool_id = tool_call["id"]
            if name not in function_map:
                self.logger.warning("Run %s-%d: Tool '%s' (id: '%s') not found in function map.", run_id, turn, name, tool_id)
                self._update_partial_response(partial_response, tool_call, ToolResult(value=f"Error: Tool {name} not found."))
                continue            
            
            args = json.loads(function["arguments"])

            func = function_map[name]
            
            if self.logger.getEffectiveLevel() <= logging.DEBUG:
                self.logger.debug("Run %s-%d: Processing tool call '%s' (id: '%s') with arguments %s", run_id, turn, name, tool_id, args)
            else:
                self.logger.info("Run %s-%d: Processing tool call '%s' (id: '%s')", run_id, turn, name, tool_id)
            t0 = time.time()
            raw_result = func(**args)
            delta_t = time.time() - t0
            if inspect.iscoroutine(raw_result):
                # Store coroutine with its metadata for parallel execution
                self.logger.info("Run %s-%d: Async tool call found: '%s' (id: '%s')", run_id, turn, name, tool_id)
                async def tool_call_wrapper(raw_result):
                    t0 = time.time()
                    ret = await raw_result
                    delta_t = time.time() - t0
                    if self.logger.getEffectiveLevel() <= logging.DEBUG:
                        self.logger.debug("Run %s-%d: (After %.2f s) Async tool call '%s' (id: '%s') returned %s", run_id, turn, delta_t, name, tool_id, ret)
                    else:
                        self.logger.info("Run %s-%d: (After %.2f s) Async tool call '%s' (id: '%s') returned successfully", run_id, turn, delta_t, name, tool_id)
                    return ret
                async_tasks.append((tool_call, tool_call_wrapper(raw_result)))
            else:
                # Handle synchronous results immediately
                if self.logger.getEffectiveLevel() <= logging.DEBUG:
                    self.logger.debug("Run %s-%d: (After %.2f s) Tool call '%s' (id: '%s') returned %s", run_id, turn, delta_t, name, tool_id, raw_result)
                else:
                    self.logger.info("Run %s-%d: (After %.2f s) Tool call '%s' (id: '%s') returned successfully", run_id, turn, delta_t, name, tool_id)
                result = await self._handle_function_result(run_id, raw_result, memory, memory_delta, caching)
                self._update_partial_response(partial_response, tool_call, result)
                if partial_response.result:
                    break

        # Execute async tasks in parallel if any exist
        if async_tasks:
            self.logger.info("Run %s-%d: Processing %d async tool call(s)", run_id, turn, len(async_tasks))
            raw_results = await asyncio.gather(*(task[1] for task in async_tasks))
            for (tool_call, _), raw_result in zip(async_tasks, raw_results):
                result = await self._handle_function_result(run_id, raw_result, memory, memory_delta, caching)
                self._update_partial_response(partial_response, tool_call, result)
                if partial_response.result:
                    break

        # TODO: Cancel all pending async tasks if the result is a final answer

        return partial_response


    def _handle_partial_response(self, run_id: str, turns: int, t0_run: float, partial_response: HandleToolCallResult, message: dict, memory: list[dict], memory_delta: list[dict]) -> Optional[Response]:
        if partial_response.filtered_tool_calls:
            # Only add tool calls to memory if there are any left after filtering
            memory_delta.append(message)
            memory_delta.extend(partial_response.messages)
        if partial_response.result:
            t_run_delta = time.time() - t0_run
            if self.logger.getEffectiveLevel() <= logging.DEBUG:
                self.logger.debug("Run %s-%d: (After %.2f s) Run completed due to final answer reached in tool call: %s", run_id, turns, t_run_delta, partial_response.result.value)
            else:
                self.logger.info("Run %s-%d: (After %.2f s) Run completed due to final answer reached in tool call", run_id, turns, t_run_delta)
            memory.extend(memory_delta)
            return Response(
                value=partial_response.result.value,
                memory_delta=memory_delta,
                agent=self,
            )
        

    def _get_response(self, run_id: str, turns: int, t0_run: float, memory: list[dict], memory_delta: list[dict]):
        memory.extend(memory_delta)
        value = self._get_value(memory[-1]["content"])
        t_run_delta = time.time() - t0_run
        if self.logger.getEffectiveLevel() <= logging.DEBUG:
            self.logger.debug("Run %s-%d: (After %.2f s) Run completed with value %s", run_id, turns, t_run_delta, value)
        else:
            self.logger.info("Run %s-%d: (After %.2f s) Run completed", run_id, turns, t_run_delta)
        return Response(
            value=value,
            memory_delta=memory_delta,
            agent=self,
        )


    def _get_user_message(self, inputs: tuple) -> dict:
        def user_message_part(input):
            if isinstance(input, str):
                return {
                    "type": "text",
                    "text": input,
                }
            elif isinstance(input, dict):
                return input
            elif isinstance(input, AnyUrl):
                # Assume image.
                return {
                    "type": "image_url",
                    "image_url": {
                        "url": str(input),
                    },
                }
            else:
                if hasattr(input, 'read'):  # is file-like object
                    file_name = input.name if input.name else None
                    mime_type = get_mime_type_from_file_like_object(input, file_name)
                    file_name = file_name or "temp_file"
                    content = input.read()
                elif isinstance(input, bytes):
                    file_name = "temp_file"
                    mime_type = get_mime_type_from_content(input)
                    content = input
                else:
                    raise ValueError(f"Unsupported element type: {type(input)}")
                base64_content = base64.b64encode(content).decode('utf-8')
                return {
                    "type": "file",
                    "file": {
                        "filename": file_name,
                        "file_data": f"data:{mime_type};base64,{base64_content}",
                    },
                }

        if len(inputs) == 1 and isinstance(inputs[0], str):
            # Keep it simple if there's only one string input
            return {"role": "user", "content": inputs[0]}
        else:
            return {"role": "user", "content": [user_message_part(input) for input in inputs]} 


    def _log_completion(self, run_id: str, turns: int, t0: float, message: dict):
        delta_t = time.time() - t0
        if self.logger.getEffectiveLevel() <= logging.DEBUG:
            self.logger.debug("Run %s-%d: (After %.2f s) Received completion: %s", run_id, turns, delta_t, message)
        else:
            if message["tool_calls"] and message["content"]:
                self.logger.info("Run %s-%d: (After %.2f s) Received completion with tool calls and text content.", run_id, turns, delta_t)
            elif message["tool_calls"]:
                self.logger.info("Run %s-%d: (After %.2f s) Received completion with tool calls.", run_id, turns, delta_t)
            elif message["content"]:
                self.logger.info("Run %s-%d: (After %.2f s) Received completion with text content.", run_id, turns, delta_t)


    async def _run_and_stream(
            self,
            run_id: str,
            memory: list[dict],
            memory_delta: list[dict],
            stream_tokens: bool,
            stream_delimiters: bool,
            stream_tool_calls: bool,
            stream_response: bool,
            max_turns: int,
            execute_tools: bool,
            caching: bool,
    ):
        t0_run = time.time()
        active_agent = self
        turns = 0

        while turns < max_turns:
            active_agent._before_chat_completion()
            message = {
                "content": "",
                "sender": active_agent.name,
                "role": "assistant",
                "function_call": None,
                "tool_calls": defaultdict(
                    lambda: {
                        "function": {"arguments": "", "name": ""},
                        "id": "",
                        "type": "",
                    }
                ),
            }

            t0 = time.time()
            # get completion with current history, agent
            completion = await active_agent._get_chat_completion(run_id, turns, memory, memory_delta, stream=True, caching=caching)

            if stream_delimiters:
                yield MessageDelimiter(delimiter=Delimiter.ASSISTANT_START, message=message)

            async for chunk in completion:
                delta = chunk.choices[0].delta.model_dump()
                if config.debug_log_streaming_deltas:
                    self.logger.debug("Run %s-%d: Received delta: %s", run_id, turns, delta)
                if delta["role"] == "assistant":
                    delta["sender"] = active_agent.name
                if "content" in delta and delta["content"]:
                    if stream_tokens:
                        yield delta["content"]
                    else:
                        yield delta
                elif "tool_calls" in delta and delta["tool_calls"]:
                    if stream_tool_calls:
                        yield delta
                else:
                    # In theory, the check for "content" and "tool_calls" should be enough, so
                    # this should never happen. However, LiteLLM seems to send some additional
                    # empty chunks. We ignore them for now.
                    pass
                delta.pop("role", None)
                delta.pop("sender", None)
                merge_chunk(message, delta)
            
            if stream_delimiters:
                yield MessageDelimiter(delimiter=Delimiter.ASSISTANT_END, message=message)

            message["tool_calls"] = list(
                message.get("tool_calls", {}).values())
            if not message["tool_calls"]:
                message["tool_calls"] = None

            active_agent._log_completion(run_id, turns, t0, message)

            if not message["tool_calls"] or not execute_tools:
                memory_delta.append(message)
                break

            # convert tool_calls from dict to objects
            tool_calls = []
            for tool_call in message["tool_calls"]:
                function = Function(
                    arguments=tool_call["function"]["arguments"],
                    name=tool_call["function"]["name"],
                )
                tool_call_object = ChatCompletionMessageToolCallParam(
                    id=tool_call["id"], function=function, type=tool_call["type"]
                )
                tool_calls.append(tool_call_object)

            # handle function calls and switching agents
            partial_response = await active_agent._handle_tool_calls(run_id, turns, tool_calls, memory, memory_delta, caching)
            
            if stream_delimiters:
                for tool_message in partial_response.messages:
                    yield MessageDelimiter(delimiter=Delimiter.TOOL_CALL, message=tool_message)
            
            response = active_agent._handle_partial_response(run_id, turns, t0_run, partial_response, message, memory, memory_delta)
            if response:
                if stream_response:
                    yield response
                else:
                    return

            if partial_response.agent:
                active_agent = partial_response.agent
            
            turns += 1

        memory.extend(memory_delta)
        if stream_response:
            yield active_agent._get_response(run_id, turns, t0_run, memory, memory_delta)
        else:
            active_agent.logger.info("Run %s-%d: (After %.2f s) Run completed", run_id, turns, time.time() - t0_run)


    async def _run(
            self,
            run_id: str,
            memory: Optional[list[dict]] = None,
            memory_delta: Optional[list[dict]] = None,
            max_turns: Optional[int] = float("inf"),
            execute_tools: Optional[bool] = True,
            caching: Optional[bool] = None,
    ) -> Response:
        t0_run = time.time()
        active_agent = self
        turns = 0

        while turns < max_turns and active_agent:
            active_agent._before_chat_completion()
            # get completion with current history, agent
            t0 = time.time()
            completion = await active_agent._get_chat_completion(run_id, turns, memory, memory_delta, caching=caching)
            message = completion.choices[0].message.model_dump()
            message["sender"] = active_agent.name

            active_agent._log_completion(run_id, turns, t0, message)

            if not message["tool_calls"] or not execute_tools:
                memory_delta.append(message)
                break

            # handle function calls and switching agents
            partial_response = await active_agent._handle_tool_calls(run_id, turns, message["tool_calls"], memory, memory_delta, caching)
            response = active_agent._handle_partial_response(run_id, turns, t0_run, partial_response, message, memory, memory_delta)
            if response:
                return response
            
            if partial_response.agent:
                active_agent = partial_response.agent

            turns += 1

        return active_agent._get_response(run_id, turns, t0_run, memory, memory_delta)


    def _get_run_id(self):
        # 6 random alphanumeric characters
        return ''.join(random.choices(string.ascii_letters + string.digits, k=6))

    
    async def run(
            self,
            *inputs,
            memory: Optional[list[dict]] = None,
            memory_delta: Optional[list[dict]] = None,
            stream: Optional[bool] = False,
            stream_tokens: bool = True,
            stream_delimiters: bool = False,
            stream_tool_calls: bool = False,
            stream_response: bool = False,
            max_turns: Optional[int] = float("inf"),
            execute_tools: Optional[bool] = True,
            caching: Optional[bool] = None,
    ) -> Response | AsyncGenerator[Response, None]:
        if memory is None:
            memory = []
        if caching is None:
            caching = config.caching

        if memory_delta is None:
            memory_delta = []
        elif memory_delta:
            raise ValueError("memory_delta must be an empty list if provided as a parameter")
        
        if inputs:
            memory_delta.append(self._get_user_message(inputs))

        run_id = self._get_run_id()

        if self.logger.getEffectiveLevel() <= logging.DEBUG:
            self.logger.debug("Run %s-0: Starting run with input(s): %s", run_id, inputs)
        else:
            self.logger.info("Run %s-0: Starting run with %d input(s)", run_id, len(inputs))

        if stream:
            return self._run_and_stream(
                run_id=run_id,
                memory=memory,
                memory_delta=memory_delta,
                stream_tokens=stream_tokens,
                stream_delimiters=stream_delimiters,
                stream_tool_calls=stream_tool_calls,
                stream_response=stream_response,
                max_turns=max_turns,
                execute_tools=execute_tools,
                caching=caching,
            )
        else:
            return await self._run(
                run_id=run_id,
                memory=memory,
                memory_delta=memory_delta,
                max_turns=max_turns,
                execute_tools=execute_tools,
                caching=caching,
            )


    def run_sync(
            self, 
            *inputs,
            memory: Optional[list[dict]] = None,
            memory_delta: Optional[list[dict]] = None,
            stream: bool = False, 
            stream_tokens: bool = True,
            stream_delimiters: bool = False,
            stream_tool_calls: bool = False,
            stream_response: bool = False,
            max_turns: int = float("inf"), 
            execute_tools: bool = True,
            caching: bool = None,
    ) -> Response:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        
        return loop.run_until_complete(
            self.run(
                *inputs, 
                memory=memory,
                memory_delta=memory_delta,
                stream=stream, 
                stream_tokens=stream_tokens, 
                stream_delimiters=stream_delimiters, 
                stream_tool_calls=stream_tool_calls, 
                stream_response=stream_response, 
                max_turns=max_turns, 
                execute_tools=execute_tools, 
                caching=caching,
            )
        )


    def apply(
            self, 
            *inputs,
            memory: Optional[list[dict]] = None,
            memory_delta: Optional[list[dict]] = None,
            stream: bool = False, 
            stream_tokens: bool = True,
            stream_delimiters: bool = False,
            stream_tool_calls: bool = False,
            stream_response: bool = False,
            max_turns: int = float("inf"), 
            execute_tools: bool = True,
            caching: bool = None,
    ) -> Response:
        """
        Synchronously apply the agent to the inputs and return the response value.
        """
        response = self.run_sync(
            *inputs,
            memory=memory,
            memory_delta=memory_delta,
            stream=stream,
            stream_tokens=stream_tokens,
            stream_delimiters=stream_delimiters,
            stream_tool_calls=stream_tool_calls,
            stream_response=stream_response,
            max_turns=max_turns,
            execute_tools=execute_tools,
            caching=caching,
        )
        if stream:
            return response
        else:
            return response.value
        

    async def __call__(
            self,
            *inputs,
            memory: Optional[list[dict]] = None,
            memory_delta: Optional[list[dict]] = None,
            stream: Optional[bool] = False,
            stream_tokens: bool = True,
            stream_delimiters: bool = False,
            stream_tool_calls: bool = False,
            stream_response: bool = False,
            max_turns: Optional[int] = float("inf"),
            execute_tools: Optional[bool] = True,
            caching: Optional[bool] = None,
    ) -> Response:
        """
        Asynchronously apply the agent to the inputs and return the response value.
        """
        response = await self.run(
            *inputs,
            memory=memory,
            memory_delta=memory_delta,
            stream=stream,
            stream_tokens=stream_tokens,
            stream_delimiters=stream_delimiters,
            stream_tool_calls=stream_tool_calls,
            stream_response=stream_response,
            max_turns=max_turns,
            execute_tools=execute_tools,
            caching=caching,
        )
        if stream:
            return response
        else:
            return response.value
        

__all__ = ["Agent", "Response", "ToolResult", "MessageDelimiter", "Delimiter"]


from slimagents.core import Agent

filename = "temp/Get_Started_With_Smallpdf.pdf"
with open(filename, "rb") as f:
    ocr = Agent(
        model="gpt-4o-mini",
        # model="anthropic/claude-3-5-sonnet-20240620",
        # model="gemini/gemini-1.5-flash",
    )
    response = ocr.run_sync(f, "Extract the text from the PDF file in markdown format. Don't include any other text or formatting.")
    print(response.value)


from pprint import pprint
from slimagents.core import Agent, ToolResult

calc_agent = Agent(
    name="Calculator",
    instructions="You are a human calculator.",
    tools=[],
    model="gpt-4o-mini",
    response_format=float,
)

def calculator():
    """
    Transfer the question to the calculator agent. Always use this tool when the user asks a question that can be answered by the calculator.
    """
    return ToolResult(agent=calc_agent, is_final_answer=True)

master = Agent(
    name="Master",
    instructions="You are a master agent that can handle all tasks.",
    tools=[calculator],
    model="gpt-4o-mini",
)

memory = []

value = master.apply("What is 2 + 2?", memory=memory)

pprint(memory)
print(value)




import asyncio
from datetime import datetime
import logging
import random
from slimagents import Agent, ToolResult
from dotenv import load_dotenv
from pydantic import BaseModel

load_dotenv()

def add(a: float, b: float) -> float:
    """Add two numbers."""
    print(f"Adding {a} and {b}")
    return a + b

def subtract(a: float, b: float) -> float:
    """Subtract two numbers."""
    print(f"Subtracting {a} and {b}")
    return a - b

def multiply(a: float, b: float) -> float:
    """Multiply two numbers."""
    print(f"Multiplying {a} and {b}")
    return a * b

def divide(a: float, b: float) -> float:
    """Divide two numbers."""
    print(f"Dividing {a} and {b}")
    return a / b

def thought(description: str, final_answer: str):
    """Always register your thought before calling another tool.
    Arguments:
        description (str): The description of your thought.
        final_answer (str): The final answer to the user's question. Leave empty if you have not reached a solution yet.
    """
    print(f"\n### Thought: {description}")
    if final_answer:
        return ToolResult(is_final_answer=final_answer)
    return "OK"

def thought(description: str, final_answer: str):
    """This is a special tool call that is used to register your reasoning for the next step, or why you have reached a solution.
    Arguments:
        description (str): The description of your thought. Describe which tool calls you will make next, or why you have reached a solution.
        final_answer (str): The final answer to the user's question. Leave empty if you have not reached a solution yet.
    """
    print(f"\n### Thought: {description}")
    if final_answer:
        print("-->", "Final answer reached:", final_answer)
        return ToolResult(value=final_answer, is_final_answer=True)
    return "OK"

def final_answer(answer: str):
    """Register your final answer."""
    print(f"--- Final answer: {answer}")
    return ToolResult(is_final_answer=answer)

def calculator(expression: str) -> float:
    """Calculate the result of an expression. IMPORTANT: Always use this tool for calculations, never try to do it yourself."""
    print(f"--- Calculating {expression}")  
    ret = eval(expression)
    print("-->", str(ret))
    return ret

def show_on_screen(text: str):
    """Show text on the screen."""
    print(f"--- Showing on screen: {text}")
    ret = "The message was shown on the screen."
    print("-->", ret)
    return ret

def get_profit_and_loss_report(month: int, year: int) -> str:
    """Get the profit and loss report for a given month and year.
    Arguments:
        month (int): The month to get the profit and loss report for. 1 - 12.
        year (int): The year to get the profit and loss report for.
    """
    print(f"--- Getting profit and loss report for {month}/{year}")
    ret = """
    | Account               | Amount  |
    |-----------------------|---------|
    | Sales                 | 1000    |
    | Cost of goods sold    | 500     |
    | Other expenses        | 300     |
    """
    print("-->", ret)
    return ret

def get_weather(location: str) -> str:
    """Get the weather for a given location."""
    print(f"--- Getting weather for {location}")
    temperature = random.randint(10, 30)
    weather = random.choice(["sunny", "cloudy", "rainy", "snowy"])
    ret = f"It's {weather} and {temperature} degrees."
    print("-->", ret)
    return ret

def get_current_date(today: str) -> str:
    """Get the current date. Always use "today" as the input."""
    print("--- Getting current date")
    ret = datetime.now().strftime("%Y-%m-%d")
    print("-->", ret)
    return ret

instructions = """
You are a helpful assistant. Your task is to come up with the next step to answer the question. 
You have access to a set of tools, and you will *always* use the tools to solve the problem. 
For each request, you will generate AT LEAST two tool calls:
1. ALWAYS generate a tool call to register your reasoning for the next step, or why you have reached a solution.
2. One or more tool calls to actually solve the problem, if you have not reached a solution yet.
3. Or a tool call to register your final answer, if you have reached a solution. ALWAYS generate this tool call before the end of the request.

IMPORTANT: In some cases the result of one tool call will be the input for the another tool call. 
In that case, only generate the first tool call. In other words, only generate tool calls that are 
independent of each other and that can be executed in any order.

REMEMBER: Always use the tool calls "thought" and "final_answer" when you have reached a solution.
"""

instructions = """
You are a helpful assistant. Your task is to come up with THE NEXT STEP to answer the user's question. 
You have access to a set of tools, and you will use the tools to solve the next step, if possible. 

In some cases the result of one or more tool calls will be the input for other tool calls. 
In that case, only generate the first tool calls. In other words, only generate tool calls that are 
independent of each other and that can be executed in any order.

If no appropriate tool call is available you can answer the question yourself from your knowledge.
IMPORTANT: Never make up any answers. If you don't know the answer, say so.
"""

instructions = """
You are a helpful assistant. Your task is to come up with THE NEXT STEP to answer the user's question. 
You have access to a set of tools, and you will use the tools to solve the next step, if possible. 

In some cases the result of one or more tool calls will be the input for other tool calls. 
In that case, only generate the first tool calls. In other words, only generate tool calls that are 
independent of each other and that can be executed in any order.

If no appropriate tool call is available you can answer the question yourself from your knowledge.
IMPORTANT: Never make up any answers. If you don't know the answer, say so.
"""

instructions = """
You are a helpful assistant. Your task is to come up with the next step to answer the incoming question. 
You have access to a set of tools, and you will always use the tools to generate the answer, if possible. 

You will ALWAYS start by generating a tool call to "thought". This is a special tool call that is used to register your reasoning for the next step, or the final answer.

You may also generate one or more tool calls to actually solve the problem, if you have not reached a solution yet.

IMPORTANT: Remember to always include the "thought" tool call"""

async def main():
    agent = Agent(
        instructions=instructions,
        # tools=[thought, final_answer, add, subtract, multiply, divide],
        # tools=[thought, final_answer, calculator, show_on_screen, get_profit_and_loss_report, get_weather],
        tools=[thought, calculator, show_on_screen, get_profit_and_loss_report, get_weather, get_current_date],
        # model="gemini/gemini-1.5-flash-002",
        model="gemini/gemini-2.0-flash-exp",
        parallel_tool_calls=None,
        # response_format=MyResult,
        # extra_llm_params={"drop_params": True},
    )

    agent.logger.setLevel(logging.DEBUG)

    # response = await agent.run("What is 3 + 3 * 4 - 2 / 2? Show the result on the screen.")
    response = await agent.run("What was the profit last month? Show the result on the screen.")
    # response = await agent.run("What is the temperature difference betweeen Oslo and Stockholm? Show the result on the screen.")
    print(response.value)


if __name__ == "__main__":
    asyncio.run(main())



# Global configuration variables
import logging

caching = False 

# The default logger for the slimagents package.
logger = logging.getLogger("slimagents")

# The root logger for the Agent class and subclasses (if separate_agent_logger is True).
# It defaults to the default logger (`slimagents`).
agent_logger = logger

# This option is used to specify if a separate logger should be created for each Agent subclass.
# If set to True, the logger will be created as a child of `agent_logger` with the name of the Agent subclass.
# If set to False, the default logger will be used.
# Default is False.
separate_agent_logger = False

# This option is used to specify if the streaming deltas should be logged. Only applicable for DEBUG level logging.
# Default is False.
debug_log_streaming_deltas = False

__all__ = ["caching", "logger", "agent_logger", "separate_agent_logger", "debug_log_streaming_deltas"]



import base64
import litellm

filename = "tests/test.pdf"
with open(filename, "rb") as f:
    file_content = f.read()
    base64_content = base64.b64encode(file_content).decode('utf-8')
    file_data = f"data:application/pdf;base64,{base64_content}"
    file = {
        "type": "file",
        "file": {
            "filename": filename,
            "file_data": file_data,
        },
    }
    text = "Extract the text from the PDF file in markdown format. Don't include any other text or formatting."
    response = litellm.completion(
        # model="gpt-4o",
        # model="anthropic/claude-3-5-sonnet-20240620",
        model="gemini/gemini-1.5-flash",
        messages=[{"role": "user", "content": [file, {"type": "text", "text": text}]}],
    )
    print(response.choices[0].message.content)
    
